/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2004 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   As a special exception, Pasquale Tricarico gives permission to
   link this program with Qt commercial edition, and distribute the
   resulting executable, without including the source code for the Qt
   commercial edition in the source distribution.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "orsa_frame.h"

#include "orsa_universe.h"
#include "orsa_config.h"

#include <iostream>

using namespace std;

namespace orsa {
  
  Frame::Frame() : vector<Body>(), UniverseTypeAwareTime() { }
  
  Frame::Frame(const double t) : vector<Body>(), UniverseTypeAwareTime(t) { }
  
  Frame::Frame(const Frame & f) : vector<Body>(f), UniverseTypeAwareTime(f) {
    
  }
  
  bool Frame::operator < (const Frame & f) const {
    return (this->UniverseTypeAwareTime::operator < (f));
  }
  
  void print(const Frame &f) {
    
    cout << "Frame time: " << f.Time() << endl;
    
    cout << "Frame size: " << f.size() << endl;
    
    unsigned int l;
    
    for (l=0;l<f.size();l++) print(f[l]);
    
  }
  
  Vector Frame::CenterOfMass() const {
    return (orsa::CenterOfMass(*this));
  }
  
  Vector Frame::Barycenter() const {
    return (orsa::CenterOfMass(*this));
  }
  
  Vector CenterOfMass(const vector<Body> & f) {
    Vector vec_sum;
    unsigned int k;
    double mass;
    double mass_sum=0;
    for (k=0;k<f.size();++k) {
      mass = f[k].mass();
      if (mass > 0.0) {
	vec_sum  += mass*f[k].position();
	mass_sum += mass;
      }
    }
    return (vec_sum/mass_sum);
  }
  
  Vector Barycenter(const vector<Body> & f) {
    return (orsa::CenterOfMass(f));
  }
  
  Vector Frame::AngularMomentum(const Vector & center) const {
    return (orsa::AngularMomentum(*this,center));
  }
  
  Vector AngularMomentum(const vector<Body> & f, const Vector & center) {
    Vector vec_sum;
    unsigned int k;
    for (k=0;k<f.size();++k) {
      if (f[k].mass() > 0) {
	vec_sum += f[k].mass()*ExternalProduct(f[k].position()-center,f[k].velocity());
      }
    }
    return (vec_sum);
  }
  
  Vector Frame::BarycentricAngularMomentum() const {
    return (orsa::BarycentricAngularMomentum(*this));
  }
  
  Vector BarycentricAngularMomentum(const vector<Body> & f) {
    return (orsa::AngularMomentum(f,Barycenter(f)));
  }
  
  double KineticEnergy(const vector<Body> & f) {
    if (f.size()==0) return (0.0);
    double energy=0.0;
    unsigned int k=0;
    while (k<f.size()) {
      energy += f[k].KineticEnergy();
      k++;
    }
    return (energy);
  }
  
} // namespace orsa
