/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2004 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   As a special exception, Pasquale Tricarico gives permission to
   link this program with Qt commercial edition, and distribute the
   resulting executable, without including the source code for the Qt
   commercial edition in the source distribution.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "orsa_interaction.h"
#include "orsa_secure_math.h"
#include "orsa_universe.h"
#include "orsa_error.h"

#include <cmath>
#include <iostream>

using namespace std;

namespace orsa {
  
  std::string label(const InteractionType it) {
    
    std::string s("");
    
    switch (it) {
    case NEWTON:                                s = "Newton";                               break;
    case NEWTON_MPI:                            s = "Newton (MPI)";                         break;
    case ARMONICOSCILLATOR:                     s = "Armonic Oscillator";                   break;
    case GALACTIC_POTENTIAL_ALLEN:              s = "Galactic Potential (Allen)";           break; 
    case GALACTIC_POTENTIAL_ALLEN_PLUS_NEWTON:  s = "Galactic Potential (Allen) + Newton";  break;
    case JPL_PLANETS_NEWTON:                    s = "JPL planets + Newton";                 break; 
    case GRAVITATIONALTREE:                     s = "Gravitational TreeCode";               break;
    case RELATIVISTIC:                          s = "Newton + Relativistic effects";        break;
    }	
    
    return s;
  }
  
  void make_new_interaction(Interaction ** i, const InteractionType type) {
    
    delete (*i);
    (*i) = 0;
    
    switch (type) {
    case NEWTON:                               (*i) = new Newton;                           break;
    case NEWTON_MPI:                        
#ifdef HAVE_MPI
      (*i) = new Newton_MPI;                  
#else
      ORSA_ERROR_MACRO("WARNING: read NEWTON_MPI interaction from application without MPI support.");
#endif
      break;
    case ARMONICOSCILLATOR:                    (*i) = new ArmonicOscillator(1,1);           break;
    case GALACTIC_POTENTIAL_ALLEN:             (*i) = new GalacticPotentialAllen;           break;
    case GALACTIC_POTENTIAL_ALLEN_PLUS_NEWTON: (*i) = new GalacticPotentialAllenPlusNewton; break;
    case JPL_PLANETS_NEWTON:                   /*************************************/      break;
    case GRAVITATIONALTREE:                    (*i) = new GravitationalTree;                break;
    case RELATIVISTIC:                         (*i) = new Relativistic;                     break;
    }    
  }
  
  // Newton
  
  Newton::Newton() : Interaction(), g(GetG()) {
    
  }
  
  Newton::Newton(const Newton &) : Interaction(), g(GetG()) {
    
  }
  
  Interaction * Newton::clone() const {
    return new Newton(*this);
  }
  
  void Newton::Acceleration(const Frame &f, vector<Vector> &a) {
    
    if (f.size() < 2) return;
    
    a.resize(f.size());
    
    unsigned int i,j;
    
    Vector d;
    
    double l;
    // double ls;    
    
    for (i=0;i<a.size();++i)
      a[i].Set(0,0,0);
    
    for (i=1;i<f.size();++i) {
      
      // cerr << "considering " << f[i].name() << "..." << endl;
      
      // if (!f[i].StillAlive()) continue;
      
      for (j=0;j<i;++j) {
	
	// cerr << "considering " << f[j].name() << "..." << endl;
	
	// if (!f[j].StillAlive()) continue;
	
	if ((f[i].mass()==0) && (f[j].mass()==0)) continue;
	
	// cerr << i << " <--> " << j << endl;
	
        // d  =  f[i].position;
        // d -=  f[j].position;
	
	d = f[i].DistanceVector(f[j]);
	
        // ls = d.LengthSquared();
        l = d.Length();
        
	// cerr << "ls = " << ls << endl;
	
        if (d.IsZero()) {
	  char err_msg[1024];
	  sprintf(err_msg,"WARNING: two objects in the same position! (%s and %s)",f[i].name().c_str(),f[j].name().c_str());
	  ORSA_ERROR_MACRO(err_msg);
	  continue;
        }
        
	// d *= secure_pow(ls,-1.5);
       	// d *= pow(ls,-1.5); // don't use the secure version for performance reasons
 	d /= l*l*l;
	
	/* a[i] -= d * f[j].mass();
	   a[j] += d * f[i].mass();  
	*/
	
	a[i] += d * f[j].mass();
        a[j] -= d * f[i].mass();
        
	// fprintf(stderr,"body %s: acc. due to %s = %g    force = %g\n",f[i].name().c_str(),f[j].name().c_str(),a[i].Length(),a[i].Length()*f[i].mass());
	
      } 
      
    } 
    
    for (i=0;i<a.size();++i) a[i] *= g;
    
  }
  
  double Newton::PotentialEnergy(const Frame &f) {
    
    double energy=0.0;
    
    if (f.size() < 2) return(0.0);
    
    unsigned int i,j;
    
    Vector d;
    
    double l;
    
    for (i=1;i<f.size();++i) {
      
      if (f[i].mass()==0) continue;
      
      for (j=0;j<i;++j) {
        
	if (f[j].mass()==0) continue;
	
	// d  =  f[i].position;
	// d -=  f[j].position;
	
	d = f[i].DistanceVector(f[j]);
	
        l = d.Length();
        
        if (d.IsZero()) {
	  char err_msg[1024];
	  sprintf(err_msg,"WARNING: two objects in the same position! (%s and %s)",f[i].name().c_str(),f[j].name().c_str());
	  ORSA_ERROR_MACRO(err_msg);
	  continue;
        }
        
	energy -= f[i].mass()*f[j].mass()/l;
	
      } 
    }  
    
    return (energy*g);
    
  }
  
  // Relativistic
  
  // Relativistic::Relativistic() : Interaction(), g(GetG()), c_2(secure_pow(GetC(),2.0)) {
  Relativistic::Relativistic() : Interaction(), g(GetG()), c_2(GetC()*GetC()) {
    
  }
  
  Relativistic::Relativistic(const Relativistic &) : Interaction(), g(GetG()), c_2(GetC()*GetC()) {
    
  }
  
  Interaction * Relativistic::clone() const {
    return new Relativistic(*this);
  }
  
  // see http://www.projectpluto.com/relativi.htm
  //
  // acc = acc_{newton} - \frac{G M}{r^3 c^2} ( (\frac{4 G M}{r} - v^2) \vec(r) + 4 (v \cdot r) \vec(v) )
  //
  void Relativistic::Acceleration(const Frame &f, vector<Vector> &a) {
    
    if (f.size() < 2) return;
    
    a.resize(f.size());
    
    unsigned int i,j;
    
    for (i=0;i<a.size();++i)
      a[i].Set(0,0,0);
    
    Vector r;
    Vector v;
    
    double r_1, r_3;
    
    for (i=1;i<f.size();++i) {
      
      for (j=0;j<i;++j) {
	
	/*** acc. on 'i' due to 'j' ***/
	
        if ((f[i].mass()==0) && (f[j].mass()==0)) continue;
	
	// 'i' - 'j'
	r = f[i].position() - f[j].position();
	v = f[i].velocity() - f[j].velocity();
	
	if (r.IsZero()) {
	  char err_msg[1024];
	  sprintf(err_msg,"WARNING: two objects in the same position! (%s and %s)",f[i].name().c_str(),f[j].name().c_str());
	  ORSA_ERROR_MACRO(err_msg);
	  continue;
	}
	
	r_1 = r.Length();
	r_3 = r.LengthSquared()*r_1;
	
	a[i] -= f[j].mass()/r_3*r + f[j].mass()/(r_3*c_2)*((4.0*g*f[j].mass()/r_1-v.LengthSquared())*r + 4.0*(r*v)*v);
	
	// now, acc. on 'j' due to 'i' so r -> -r and v -> -v ***/
	//
	a[j] += f[i].mass()/r_3*r + f[i].mass()/(r_3*c_2)*((4.0*g*f[i].mass()/r_1-v.LengthSquared())*r + 4.0*(r*v)*v);
	
      } 
      
    } 
    
    for (i=0;i<a.size();++i) a[i] *= g;
  }
  
  double Relativistic::PotentialEnergy(const Frame &f) {
    Newton newton;
    return (newton.PotentialEnergy(f));
  }
  
  // ArmonicOscillator
  
  ArmonicOscillator::ArmonicOscillator(const double free_length_in, const double k_in) : Interaction(), free_length(free_length_in), k(k_in) {
    
  }
  
  ArmonicOscillator::ArmonicOscillator(const ArmonicOscillator & i) : Interaction(), free_length(i.free_length), k(i.k) {
    
  }
  
  Interaction * ArmonicOscillator::clone() const {
    return new ArmonicOscillator(*this);
  }
  
  void ArmonicOscillator::Acceleration(const Frame& f, vector<Vector>& a) {
    
    if (f.size() < 2) return;
    
    a.resize(f.size());
    
    unsigned int i,j;
    
    Vector d, da;

    double ls;
    
    for (i=0;i<f.size();++i)
      a[i].Set(0,0,0);
    
    for (i=1;i<f.size();++i) {
      
      if (f[i].mass()==0) continue;
      
      for (j=0;j<i;++j) {
        
	// if ((f[i].mass==0) || (f[j].mass==0)) continue;
	
        // assert(i!=j);
        
	// d  =  f[i].position;
	// d -=  f[j].position;
	
	d = f[i].DistanceVector(f[j]);

	ls = d.Length();
        
        if (d.IsZero()) {
	  char err_msg[1024];
	  sprintf(err_msg,"WARNING: two objects in the same position! (%s and %s)",f[i].name().c_str(),f[j].name().c_str());
	  ORSA_ERROR_MACRO(err_msg);
	  continue;
        }
        
	da = d*(ls-free_length)/ls;
	
	a[i] += da;
        a[j] -= da;
        
      } 
    }  
    
    for (i=0;i<a.size();++i) 
      if (f[i].mass() != 0) a[i] *= k/f[i].mass();
    
  }
  
  double ArmonicOscillator::PotentialEnergy(const Frame &f) {
    
    double energy=0.0;
    
    if (f.size() < 2) return(0.0);
    
    unsigned int i,j;
    
    Vector d;
    
    double ls;
    
    for (i=1;i<f.size();++i) {
      
      if (f[i].mass()==0) continue;
      
      for (j=0;j<i;++j) {
        
	// if ((f[i].mass==0) && (f[j].mass==0)) continue;
	
	// d  =  f[i].position;
	// d -=  f[j].position;
	
	d = f[i].DistanceVector(f[j]);
	
        if (d.IsZero()) {
	  char err_msg[1024];
	  sprintf(err_msg,"WARNING: two objects in the same position! (%s and %s)",f[i].name().c_str(),f[j].name().c_str());
	  ORSA_ERROR_MACRO(err_msg);
	  continue;
        }
        
	// ls = secure_pow(d.Length()-free_length,2);
	ls = pow(d.Length()-free_length,2); // don't use the secure version for performance reasons
	
	energy += ls/2.0;
	
      } 
    }  
    
    return (energy*k);
    
  }
  
  // GalacticPotentialAllen
  // see: http://adsabs.harvard.edu/cgi-bin/nph-bib_query?bibcode=1991RMxAA..22..255A&db_key=AST&high=3e8aa73e3d07980
  /* @ARTICLE{1991RMxAA..22..255A,
     author = {{Allen}, C. and {Santillan}, A.},
     title = "{An improved model of the galactic mass distribution for orbit computations}",
     journal = {Revista Mexicana de Astronomia y Astrofisica},
     year = 1991,
     month = oct,
     volume = 22,
     pages = {255-263},
     adsurl = {http://adsabs.harvard.edu/cgi-bin/nph-bib_query?bibcode=1991RMxAA..22..255A&db_key=AST},
     adsnote = {Provided by the NASA Astrophysics Data System}
     }
  */
  
  GalacticPotentialAllen::GalacticPotentialAllen() : Interaction() {
    
    g = GetG();
    
    mb = FromUnits(1.40592e10,   MSUN);
    bb = FromUnits(0.3873e3,     PARSEC);
    md = FromUnits(8.5608e10,    MSUN);     
    ad = FromUnits(5.3178e3,     PARSEC);     
    bd = FromUnits(0.25e3,       PARSEC);
    mh = FromUnits(10.3836745e10,MSUN);
    ah = FromUnits(12.0e3,       PARSEC); 
    
  }
  
  
  GalacticPotentialAllen::GalacticPotentialAllen(const GalacticPotentialAllen &) : Interaction() {
    
    g = GetG();
    
    mb = FromUnits(1.40592e10,   MSUN);
    bb = FromUnits(0.3873e3,     PARSEC);
    md = FromUnits(8.5608e10,    MSUN);     
    ad = FromUnits(5.3178e3,     PARSEC);     
    bd = FromUnits(0.25e3,       PARSEC);
    mh = FromUnits(10.3836745e10,MSUN);
    ah = FromUnits(12.0e3,       PARSEC); 
    
  }
  
  Interaction * GalacticPotentialAllen::clone() const {
    return new GalacticPotentialAllen(*this);
  }
  
  void GalacticPotentialAllen::Acceleration(const Frame &f, vector<Vector> &a) {
    
    a.resize(f.size());
    
    unsigned int i;
    for (i=0;i<a.size();++i) a[i].Set(0,0,0);
    
    double r2,z2,r,rho;
    double fbr,fdr,fhr,fhrc;
    double fbz,fdz,fhz,fhzc;
    double fr,fz;
    Vector x;
    
    for (i=0;i<f.size();++i) {
      
      // if (!f[i].StillAlive()) continue;
      
      x = f[i].position();
      
      // auxiliary components
      r2  = x.x*x.x+x.y*x.y;
      z2  = x.z*x.z;
      r   = sqrt(r2);
      rho = x.Length();
      
      // accel. components along the r direction
      fbr  = -(mb*r/(secure_pow(r2+z2+bb*bb,1.5)));
      fdr  = -(md*r/(secure_pow(r2+secure_pow(ad+sqrt(z2+bd*bd),2),1.5)));
      fhr  = 1.02*mh*r/(ah*ah*secure_pow(1+1.0/secure_pow(rho/ah,1.02),2)*secure_pow(rho/ah,2.02)*rho);
      fhrc = -(mh/(1.02*ah)*(1.0404*(r*secure_pow(rho/ah,0.02))/(ah*rho*secure_pow(1.0+secure_pow(rho/ah,1.02),2))+1.02*(r*secure_pow(rho/ah,0.02))/(ah*rho*(1.0+secure_pow(rho/ah,1.02)))));
      
      // accel. components along the z direction
      fbz  = -(mb*x.z/(secure_pow(r2+z2+bb*bb,1.5)));
      fdz  = -(md*x.z*(ad+sqrt(z2+bd*bd))/(sqrt(z2+bd*bd)*secure_pow(r2+secure_pow(ad+sqrt(z2+bd*bd),2),1.5)));
      fhz  = 1.02*mh*x.z/(ah*ah*secure_pow(1.0+1.0/secure_pow(rho/ah,1.02),2)*secure_pow(rho/ah,2.02)*rho);
      fhzc = -(mh/(1.02*ah)*(1.0404*(x.z*secure_pow(rho/ah,0.02))/(ah*rho*secure_pow(1.0+secure_pow(rho/ah,1.02),2))+1.02*(x.z*secure_pow(rho/ah,0.02))/(ah*rho*(1.0+secure_pow(rho/ah,1.02)))));
      
      fr=fbr+fdr+fhr+fhrc;
      fz=fbz+fdz+fhz+fhzc;
      
      a[i].x = fr*x.x/r;
      a[i].y = fr*x.y/r;
      a[i].z = fz;
    } 
    
    for (i=0;i<a.size();++i) a[i] *= g;
    
    // for (i=0;i<a.size();++i) printf("a[%i] = %g %g %g\n",i,a[i].x,a[i].y,a[i].z);
    
  }
  
  double GalacticPotentialAllen::PotentialEnergy(const Frame &f) {
    
    // to be tested...
    
    double energy = 0.0;
    
    double pb,pd,ph;
    double r2,z2,r,rho;
    unsigned int i;
    Vector x;
    
    for (i=0;i<f.size();++i) {
      
      x = f[i].position();
      
      // auxiliary components
      r2  = x.x*x.x+x.y*x.y;
      z2  = x.z*x.z;
      r   = sqrt(r2);
      rho = x.Length();
      
      pb = mb/sqrt(r2+z2+bb*bb);
      pd = md/sqrt(r2+secure_pow(ad+sqrt(z2+bd*bd),2));
      ph = (mh/(rho/ah))*(secure_pow(rho/ah,2.02))/(1.0+secure_pow(rho/ah,1.02));
      
      energy -= (pb+pd+ph);
      
    } 
    
    return (energy*g);
  }
  
  // JPLPlanetsNewton
  
  JPLPlanetsNewton::JPLPlanetsNewton(list<JPL_planets> & l_in) : Interaction(), l(l_in) {
    if (universe->GetUniverseType() != Real) {
      cerr << "error: using the JPLPlanetsNewton interaction in a non-Real universe!" << endl;
      exit(0);
    }
    g = GetG();
  }
  
  JPLPlanetsNewton::JPLPlanetsNewton(const JPLPlanetsNewton & i) : Interaction(), l(i.l) {
    if (universe->GetUniverseType() != Real) {
      cerr << "error: using the JPLPlanetsNewton interaction in a non-Real universe!" << endl;
      exit(0);
    }
    g = GetG();
  }
  
  Interaction * JPLPlanetsNewton::clone() const {
    return new JPLPlanetsNewton(*this);
  }
  
  void JPLPlanetsNewton::Acceleration(const Frame & f, vector<Vector> & a) {
    
    a.resize(f.size());
    
    SetupSolarSystem(planets_frame,l,f);
    
    for (unsigned int i=0;i<a.size();++i)
      a[i].Set(0,0,0);
    
    for (unsigned int i=0;i<f.size();++i) {
      if (f[i].mass() > 0) {
	ORSA_ERROR_MACRO("ERROR: using the JPLPlanetsNewton interaction with massive objects!");
	// exit(0);
	return;
      }
    }
    
    Vector d;
    
    double l;
    
    for (unsigned int i=0;i<f.size();++i) {
      
      // if (!f[i].StillAlive()) continue;
      
      for (unsigned int j=0;j<planets_frame.size();++j) {
	
	// d  =              f[i].position;
	// d -=  planets_frame[j].position;
	
	d = f[i].DistanceVector(planets_frame[j]);
	
	// ls = d.LengthSquared();
	l = d.Length();
	
        if (d.IsZero()) {
	  char err_msg[1024];
	  sprintf(err_msg,"WARNING: two objects in the same position! (%s and %s)",
		  f[i].name().c_str(),
		  planets_frame[j].name().c_str());
	  ORSA_ERROR_MACRO(err_msg);
	  continue;
        }
        
	// d *= secure_pow(ls,-1.5);
	d /= l*l*l;
	
	a[i] += d * planets_frame[j].mass();
	
      }
    }
    
    for (unsigned int i=0;i<a.size();++i) a[i] *= g;
  }
  
  double JPLPlanetsNewton::PotentialEnergy(const Frame & f) {
    SetupSolarSystem(planets_frame,l,f);
    return newton_itg.PotentialEnergy(planets_frame);
  }
  
} // namespace orsa
