/* 
   ORSA - Orbit Reconstruction, Simulation and Analysis
   Copyright (C) 2002-2004 Pasquale Tricarico
   
   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License
   as published by the Free Software Foundation; either version 2
   of the License, or (at your option) any later version.
   
   As a special exception, Pasquale Tricarico gives permission to
   link this program with Qt commercial edition, and distribute the
   resulting executable, without including the source code for the Qt
   commercial edition in the source distribution.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#include "orsa_universe.h"
#include "orsa_file.h"
#include "orsa_error.h"

#include <iostream>
#include <string>
#include <limits>

using namespace std;

namespace orsa {
  
  // global units
  Units * units = 0;
  
  // active universe
  Universe * universe = 0;
  
  // active configuration
  Config * config = 0;
  
  // global JPL file
  JPLFile * jpl_file = 0;
  
  // JPL cache global class
  JPLCache * jpl_cache = 0;
  
  // world visible LocationFile
  LocationFile * location_file = 0;
  
  Universe::Universe() : std::vector<Evolution*>(), type(Simulated), sys(ECLIPTIC), timescale(ET) {
    common_init(AU,MSUN,YEAR);
  }
  
  Universe::Universe(length_unit lu, mass_unit mu, time_unit tu, UniverseType ut, ReferenceSystem rs, TimeScale ts) : std::vector<Evolution*>(), type(ut), sys(rs), timescale(ts) {
    common_init(lu,mu,tu);
  }
  
  void Universe::common_init(const length_unit lu, const mass_unit mu, const time_unit tu) {
    delete universe; 
    universe = 0;
    
    if (!orsa_paths) orsa_paths = new OrsaPaths;
    if (!debug) debug = new Debug;
    //
    if (!config) {
      config = new Config;
    }
    config->read_from_file();
    //
    if (!units) {
      units = new Units;
      units->SetSystem(tu,lu,mu);
    }
    //
    if (!jpl_file) {
      jpl_file = new JPLFile(config->paths[JPL_EPHEM_FILE]->GetValue().c_str());
    }
    //
    if (!jpl_cache) {
      jpl_cache = new JPLCache;
    }
    //
    if (!location_file) {
      location_file = new LocationFile; 
      location_file->SetFileName(config->paths[OBSCODE]->GetValue().c_str());
      location_file->Open();
      location_file->Read();
      location_file->Close();
    }
    //
    modified = true;
    default_Date_timescale = timescale;
    universe = this;
  }
  
  Universe::~Universe() {
    /* 
       int k;
       k = size();
       while (k>0) {
       --k;
       if ((*this)[k]) (*this)[k]->clear();
       }
    */
    // NOTE: keep these two loops separate!
    int k = size();
    while (k>0) {
      --k;
      delete (*this)[k];
      (*this)[k] = 0;
    }
    
    universe = 0;
  }
  
  // declared in orsa_units.h -- to put elsewhere in future
  double GetG()        { return (orsa::units->GetG()); }
  double GetG_MKS()    { return (orsa::units->GetG_MKS()); }
  double GetMSun()     { return (orsa::units->GetMSun()); }
  double GetC()        { return (orsa::units->GetC()); }
  
  double FromUnits(double x, time_unit   u, double power) { return orsa::units->FromUnits(x,u,power); };
  double FromUnits(double x, length_unit u, double power) { return orsa::units->FromUnits(x,u,power); };
  double FromUnits(double x, mass_unit   u, double power) { return orsa::units->FromUnits(x,u,power); };
  
  string TimeLabel()   { return orsa::units->TimeLabel();   };
  string LengthLabel() { return orsa::units->LengthLabel(); };
  string MassLabel()   { return orsa::units->MassLabel();   };
  
  unsigned int Evolution::used_evolution_id = 0;
  
  Evolution::Evolution() : std::vector<Frame>(), id(used_evolution_id++) {
    sample_period = 0;
    integrator    = 0;
    interaction   = 0;
    max_unsaved_substeps_active = false;
    _integrating=false;
  }
  
  //! This copy constructor copies the parameters, not the frames!
  Evolution::Evolution(const Evolution & e) : std::vector<Frame>(), id(used_evolution_id++) {
    sample_period = e.sample_period;
    integrator    = e.integrator->clone();
    interaction   = e.interaction->clone();
    //
    max_unsaved_substeps_active = false;
    _integrating=false;
  }
  
  Evolution::~Evolution() {
    delete integrator;
    integrator = 0;
    delete interaction;
    interaction = 0;
  }
  
  void Evolution::SetIntegrator(const IntegratorType type) {
    make_new_integrator(&integrator,type);
  }
  
  void Evolution::SetIntegrator(const Integrator * itg) {
    delete integrator;
    integrator = itg->clone();
  }
  
  void Evolution::SetIntegratorTimeStep(const UniverseTypeAwareTimeStep ts) {
    integrator->timestep = ts;
  }
  
  void Evolution::SetIntegratorAccuracy(const double a) {
    integrator->accuracy = a;
  }
  
  void Evolution::SetInteraction(const InteractionType type) {
    make_new_interaction(&interaction,type);
  }
  
  void Evolution::SetInteraction(const Interaction * itr) {
    delete interaction;
    interaction = itr->clone();
  }
  
  void Evolution::SetSamplePeriod(const UniverseTypeAwareTimeStep sp) {
    sample_period = sp;
  }
  
  void Evolution::Integrate(const UniverseTypeAwareTime &time_stop, const bool save_last_anyway) {
    
    if (integrator==0) {
      ORSA_ERROR_MACRO("WARNING! Integrator not initialized!");
      return;
    }
    
    if (interaction==0) {
      ORSA_ERROR_MACRO("WARNING! Interaction not initialized!");
      return;
    }
    
    integration_started();
    
    // FIXME: if the timestep is not set?
    // if (sample_period == 0) sample_period = integrator->timestep;
    if (sample_period.IsZero()) sample_period = integrator->timestep;
    sample_period = sample_period.absolute();
    
    Frame f_start,f_stop;
    
    if (size() == 0) {
      ORSA_ERROR_MACRO("ERROR: no starting frame in integration.");
      return;
    }
    
    f_start = f_stop = (*this)[size()-1];
    
    const double time_start = f_start.Time();
    
    bool save_frame = false;
    
    bool continue_integration = true;
    
    // double sign;
    int sign;
    
    // if ((time_stop.Time()-f_stop.Time()) > 0) {
    if ((time_stop - f_stop) > 0) {
      // integrator->timestep = fabs(integrator->timestep);
      integrator->timestep = integrator->timestep.absolute();
      sign = +1;
    } else {
      // integrator->timestep = -fabs(integrator->timestep);
      integrator->timestep = -integrator->timestep.absolute();
      sign = -1;
    }
    
    // important init value!
    // double last_genuine_timestep = integrator->timestep;
    UniverseTypeAwareTimeStep last_genuine_timestep = integrator->timestep;
    
    unsigned int unsaved_substeps = 0;
    unsigned int total_substeps = 0;
    
    while ( (continue_integration) && (((time_stop.Time()-f_stop.Time())/(time_stop.Time()-time_start)) > 0) ) {
      
      f_start = f_stop;
      
      /* 
	 fprintf(stderr,
	 "==========================         \n"
	 "f_start.Time()           : %20.12e \n"
	 "integrator->timestep     : %20.12e \n"
	 "(*this)[size()-1].Time() : %20.12e \n"
	 "sample_period            : %20.12e \n"
	 "frames                   : %zi     \n"
	 "==========================         \n",
	 f_start.Time(),integrator->timestep,(*this)[size()-1].Time(),sample_period,size()
	 );
      */
      /* 
	 { 
	 char err_msg[1024];
	 sprintf(err_msg,
	 "\n"
	 "==========================         \n"
	 "f_start.Time()           : %20.12e \n"
	 "integrator->timestep     : %20.12e \n"
	 "(*this)[size()-1].Time() : %20.12e \n"
	 "sample_period            : %20.12e \n"
	 "frames                   : %zi     \n"
	 "==========================         \n",
	 f_start.Time(),integrator->timestep.GetDouble(),(*this)[size()-1].Time(),sample_period.GetDouble(),size());
	 ORSA_ERROR_MACRO(err_msg);   
	 }
      */
      
      // if ( fabs(f_start.Time() + integrator->timestep - (*this)[size()-1].Time()) > sample_period)  {
      // if ( (f_start + integrator->timestep - (*this)[size()-1]).absolute() > sample_period) {
      if ( (f_start - (*this)[size()-1] + integrator->timestep).absolute() > sample_period) {
	last_genuine_timestep = integrator->timestep;
	// integrator->timestep = (*this)[size()-1].Time() + sign*sample_period - f_start.Time();
    	integrator->timestep = (*this)[size()-1] - f_start + sign*sample_period;
	
	/* 
	   fprintf(stderr,
	   "timestep changed: from %.20e to %.20e      val: %.20e sp: %.20e\n",
	   last_genuine_timestep.GetDouble(),
	   integrator->timestep.GetDouble(),
	   (f_start + integrator->timestep - (*this)[size()-1]).absolute().GetDouble(),
	   sample_period.GetDouble());
	*/
      }
      
      if (integrator->timestep > 0) {
	// cerr << "timestep > 0 !!!" << endl;
	// if ( (f_start.Time()+integrator->timestep) > (time_stop.Time()) ) {
	if ( (f_start+integrator->timestep) > time_stop) {
	  // integrator->timestep = time_stop.Time() - f_start.Time();
	  integrator->timestep = time_stop - f_start;
	}
      } else if (integrator->timestep < 0) {
	// cerr << "timestep < 0 !!!" << endl;
	// if ( (f_start.Time()+integrator->timestep) < (time_stop.Time()) ) {
	if ( (f_start+integrator->timestep) < time_stop) {
	  // integrator->timestep = time_stop.Time() - f_start.Time();
	  integrator->timestep = time_stop - f_start;
	}
      } else {
	ORSA_ERROR_MACRO("WARNING! Timestep equal to zero!");
	// timestep == 0...
      }
      
      // step
      if (integrator->timestep.IsZero()) {
	ORSA_ERROR_MACRO("zero timestep, no integrator->Step() call...");
      } else {
	integrator->Step(f_start,f_stop,interaction);
	if (f_start.Time() != f_stop.Time()) {
	  ++total_substeps;
	}
      }
      
      /* 
	 {
	 // debug
	 fprintf(stderr,"..after step: %g\n",((f_stop - (*this)[size()-1]).absolute() - sample_period).absolute().GetDouble());
	 fprintf(stderr,"....timestep: %g\n",integrator->timestep.GetDouble());
	 fprintf(stderr,"....time....: %g\n",f_stop.Time());
	 }
      */
      
      /* 
	 { 
	 // debug
	 fprintf(stderr,
	 "save frame test: %.20e\n"
	 "sign*sample_period: %.20e      (*this)[size()-1].GetTime(): %.9f\n",
	 // ((f_stop-(*this)[size()-1]).absolute()-sample_period).absolute().GetDouble());
	 ( (*this)[size()-1] - f_stop + sign*sample_period).GetDouble(),
	 sign*sample_period.GetDouble(),
	 (*this)[size()-1].GetTime());
	 }
      */
      
      // save frame
      switch (universe->GetUniverseType()) {
      case Real:
	if ( ( (*this)[size()-1] - f_stop + sign*sample_period).IsZero() ) { 
	  save_frame = true;
	}
	break;
      case Simulated:
	if ( fabs( fabs(f_stop.Time() - (*this)[size()-1].Time()) - sample_period.GetDouble()) < (fabs(f_stop.Time())+fabs((*this)[size()-1].Time())+fabs(sample_period.GetDouble()))*total_substeps*std::numeric_limits<double>::epsilon()) {
	  save_frame = true;
	}
	break;
      }
      
      // end of integration
      /* if ( fabs(f_stop.Time() - time_stop.Time()) < fabs(time_stop.Time())*std::numeric_limits<double>::epsilon()) {
	 continue_integration=false;
	 if (save_last_anyway) save_frame=true;
	 }
      */
      //
      // end of integration
      switch (universe->GetUniverseType()) {
      case Real:
	if (f_stop == time_stop) {
	  continue_integration = false;
	  if (save_last_anyway) save_frame = true;
	}
	break;
      case Simulated:
	if ( fabs(f_stop.Time() - time_stop.Time()) < fabs(time_stop.Time())*total_substeps*std::numeric_limits<double>::epsilon()) {
	  continue_integration = false;
	  if (save_last_anyway) save_frame = true;
	}
	break;
      }
      
      
      if (save_frame) {
	// step_done(time_start,time_stop.Time(),last_genuine_timestep,f_stop,continue_integration);
	step_done(time_start,time_stop,last_genuine_timestep,f_stop,continue_integration);
	push_back(f_stop);
	save_frame = false;
	integrator->timestep = last_genuine_timestep;
	unsaved_substeps = 0;
      } else {
	++unsaved_substeps;
	if (max_unsaved_substeps_active) {
	  if (unsaved_substeps > max_unsaved_substeps) {
	    continue_integration=false;
	    ORSA_ERROR_MACRO("WARNING!!! max unsaved substeps hit!!! stopping integration!");
	  }
	}
      }
    }
    
    integration_finished();
  }
  
  void Evolution::Integrate(const Frame & f, const UniverseTypeAwareTime & utat_start, const UniverseTypeAwareTime& utat_stop) {
    
    // clear Evolution
    clear();
    
    // double t_start=utat_start.Time(), t_stop=utat_stop.Time();
    UniverseTypeAwareTime t_start = utat_start;
    UniverseTypeAwareTime t_stop  = utat_stop;
    /* 
       if (utat_start.Time() > utat_stop.Time()) {
       t_start = utat_stop.Time();
       t_stop  = utat_start.Time();
       }
    */
    
    if (utat_start > utat_stop) {
      t_start = utat_stop;
      t_stop  = utat_start;
    }
    
    Frame running_frame;
    
    if (t_start < f) {
      if (t_stop < f) {
	push_back(f);
	Integrate(t_stop);
	running_frame = (*this)[size()-1];
      	clear();
	push_back(running_frame);
	Integrate(t_start);
      } else {
	push_back(f);
	Integrate(t_start);
	// swap
	(*this)[0] = (*this)[size()-1];
	(*this)[size()-1] = f;
	//
	Integrate(t_stop);
      }
    } else {
      push_back(f);
      Integrate(t_start);
      running_frame = (*this)[size()-1];
      clear();
      push_back(running_frame);
      Integrate(t_stop);
    }
    
    sort(begin(),end());
  }
  
  // Frame StartFrame(const vector<BodyWithEpoch> &f, vector<JPLBody> &jf, Interaction *itr, Integrator *itg, const UniverseTypeAwareTime &t) {
  Frame StartFrame(const vector<BodyWithEpoch> & f, vector<JPL_planets> & jf, const Interaction * itr, const Integrator * itg, const UniverseTypeAwareTime & t) {
  // Frame StartFrame(const vector<BodyWithEpoch> & f, vector<JPL_planets> & jf, InteractionType interaction_type, IntegratorType integrator_type, const UniverseTypeAwareTime & t) {
    
    // const double original_timestep = itg->timestep;
    const UniverseTypeAwareTimeStep original_timestep = itg->timestep;
    
    Frame frame_start;
    
    switch (universe->GetUniverseType()) {
    case Real: { 
      frame_start.SetDate(t.GetDate());
      Frame          running_frame;
      Evolution      running_evolution;
      // running_evolution.interaction = itr->clone();
      // running_evolution.integrator  = itg->clone();
      running_evolution.SetIntegrator(itg);
      running_evolution.SetInteraction(itr);
      
      // put the JPL bodies into the frame_start
      /* 
	 if (jf.size()) {
	 Date ddf = t.GetDate();
	 unsigned int k;
	 for (k=0;k<jf.size();++k) {
	 jf[k].SetEpoch(ddf);
	 Body b(jf[k].name(),
	 jf[k].mass(),
	 jf[k].radius(),
	 jf[k].position(),
	 jf[k].velocity());
	 frame_start.push_back(b);
	 }
	 }
      */
      //
      if (jf.size()) {
	unsigned int k;
	for (k=0;k<jf.size();++k) {
	  frame_start.push_back(JPLBody(jf[k],t.GetDate()));
	}
      }
      
      if (f.size()) {
	
	Frame last_frame;
	unsigned int base_bodies, l;
	
	unsigned int j=0;
      	while (j < f.size()) {
	  
	  running_frame.clear();
	  // running_frame.SetDate(f[j].epoch.GetDate());
	  running_frame.SetDate(f[j].GetEpoch().GetDate());
	  
	  running_frame.push_back(f[j]);
	  ++j;
	  
	  // while ((j<f.size()) && (f[j].epoch.GetDate() == running_frame.GetDate())) {
	  while ((j<f.size()) && (f[j].Epoch().GetDate() == running_frame.GetDate())) {
	    running_frame.push_back(f[j]);
	    ++j;
	  }
	  
	  base_bodies = running_frame.size();
	  
	  // JPL 
	  /* 
	     Date ddt = running_frame.GetDate();
	     unsigned int k;
	     for (k=0;k<jf.size();++k) {
	     jf[k].SetEpoch(ddt);
	     Body b(jf[k].name(),
	     jf[k].mass(),
	     jf[k].radius(),
	     jf[k].position(),
	     jf[k].velocity());
	     running_frame.push_back(b);
	     }
	  */
	  {
	    unsigned int k;
	    for (k=0;k<jf.size();++k) {
	      running_frame.push_back(JPLBody(jf[k],running_frame.GetDate()));
	    }
	  }
	  
	  running_evolution.clear();
	  running_evolution.push_back(running_frame);
	  // running_evolution.sample_period = fabs(running_frame.GetTime() - frame_start.GetTime());
	  running_evolution.SetSamplePeriod((running_frame - frame_start).absolute());
	  running_evolution.SetIntegratorTimeStep(original_timestep);
	  running_evolution.Integrate(t);
	  
	  last_frame = running_evolution[running_evolution.size()-1];
	  if (fabs(last_frame.Time()-frame_start.Time()) > FromUnits(1.0e-3,SECOND)) {
	    /* 
	       fprintf(stderr,"!!! last_frame.Time() != frame_start.Time() --->> %30.20f != %30.20f\n",last_frame.Time(),frame_start.Time());
	       cerr << "|dT| = " << FromUnits(fabs(last_frame.Time()-frame_start.Time()),SECOND,-1) << " seconds" << endl;
	       print(last_frame);
	    */
	    char err_msg[1024];
	    sprintf(err_msg,
		    "Warning: last_frame.Time() != frame_start.Time() --->> %30.20f != %30.20f\n"
		    "|dT| = %g seconds\n",
		    last_frame.Time(),frame_start.Time(),FromUnits(fabs(last_frame.Time()-frame_start.Time()),SECOND,-1));
	    strcat(err_msg,"objects in frame: ");
	    unsigned int k;
	    for (k=0;k<last_frame.size();++k) {
	      strcat(err_msg,last_frame[k].name().c_str());
	      strcat(err_msg,", ");
	    }
	    ORSA_ERROR_MACRO(err_msg);
	  }
	  
	  for (l=0;l<base_bodies;++l) {
	    frame_start.push_back(last_frame[l]);
	  }
	}
      }
      break;
    }
    case Simulated: {
      frame_start.SetTime(0.0);
      unsigned int k;  
      for (k=0;k<f.size();++k) {
	frame_start.push_back(f[k]);
      }
      break;
    }
    }
    
    return frame_start;
  }
  
} // namespace orsa
